<?php
/**
 * Common daemons functions
 *
 * PHP version 5
 *
 * LICENSE: This source file is subject to version 3.0 of the PHP license
 * that is available through the world-wide-web at the following URI:
 * http://www.php.net/license/3_0.txt.  If you did not receive a copy of
 * the PHP License and are unable to obtain it through the web, please
 * send a note to license@php.net so we can mail you a copy immediately.
 *
 * @category   NA
 * @package    NA
 * @author     Nicolas Bruley / Peer 2 World <contact@weezo.net>
 * @copyright  2005-2008 Nicolas Bruley / Peer 2 World
 * @license    http://www.php.net/license/3_0.txt  PHP License 3.0
 * @version    CVS: $Id:$
 * @link       http://www.weezo.net
 * @since      File available since Release 2.1.0H
 */


/**
 * @desc Daemons management class
 *
 */
class WDaemons{
	/**
	 * @desc Create & launch a new daemin
	 *
	 * @param string $name: daemon's script name (without path and .php)
	 * @param integer $maxUnrespondingTime: max time (s) after wich daemon should be restarted if not responding (0 not to check)
	 * @param integer $lifetime: time (s) after which daemon is restarted (0 for infinite)
	 * @return WDaemon created daemon object
	 */
    static function create($name,$maxUnrespondingTime=0,$lifetime=0){
    	// Create daemon in daemons list
    	$daemonsList=cfMGetVar('daemonsList');
    	$daemonsList[$name]=array('maxUnrespondingTime'=>$maxUnrespondingTime,'lifetime'=>$lifetime);
    	cfMSetVar('daemonsList',$daemonsList);
    	unset($daemonsList);

    	// Start daemon
    	$d=new WDaemon($name); $d->start();
    	return $d;
    }


    /**
     * @desc Destroy a daemon
     *
     * @param string $name
     */
    static function destroy($name){
    	// Remove from list
    	$daemonsList=cfMGetVar('daemonsList');
    	if(isset($daemonsList[$name])){
    		unset($daemonsList[$name]);
    		cfMSetVar('daemonsList',$daemonsList);
    	}

    	// Destroy daemon's PID so it actually exit, and all other in-memory data
    	cfMUnsetVar('daemonPID'.$name);
    	cfMUnsetVar('daemonStatus'.$name);
    }

   /**
    * @desc Return a deamon
    *
    * @param string $name
    * @return WDaemon
    */
    static function daemon($name){return new WDaemon($name);}

    /**
     * @desc Monitor and restart hanged or too old daemons
     *
     */
	static function monitor(){
    	foreach (cfMGetVar('daemonsList') as $daemonName=>$foo){
    		// Don't monitor main daemon as this one monitor others
    		if($daemonName=='main') continue;
    		$d=WDaemons::daemon($daemonName);

    		//if(!$d->isAlive()) cfDebugSingle($daemonName.' is dead'); else cfDebugSingle($daemonName.' is alive');

    		if(!$d->isAlive()/* || $d->isTooOld()*/) $d->restart();
    	}
	}

	/**
	 * @desc Return daemons definitions array
	 *
	 * @return array
	 */
	static function getList(){return cfMGetVar('daemonsList');}

	/**
	 * @desc Return daemons definitions array
	 *
	 * @return array
	 */
	static function dump(){
		foreach (cfMGetVar('daemonsList') as $name=>$def) {
			echo '<b>'.$name.'</b>&nbsp;';
			@list($ping,$memory,$startTime)=cfMGetVar('daemonStatus'.$name);
			echo ' Start time:'.date('H:i:s',$startTime).' - Last ping: '.(time()-$ping).'s ago (max='.$def['maxUnrespondingTime'].'s) - Memory: '.number_format($memory).'<a href="/local/daemons/'.$name.'.php?stop" style="margin-left:3em">Stop</a><br>';
		}
	}
}


/**
 * Daemon object
 *
 */
class WDaemon{
	private $name;			// Daemon's name
	private $definition;	// Daemons definition array
	private $ok=false;		// True if daemon exists

	/**
	 * @desc Constructor
	 *
	 * @param string $name
	 */
	function __construct($name){
		//cfDbg('construct daemon '.$name.' in '.$_SERVER['REQUEST_URI'].((isset($daemonsList[$name]))?' but daemon already exists':' and create'));
		$this->name=$name;
		$daemonsList=cfMGetVar('daemonsList');
		if(isset($daemonsList[$name])) {
			$this->definition=$daemonsList[$name];
			$this->ok=true;
		}
		else $this->definition=array('maxUnrespondingTime'=>0,'lifetime'=>0);
	}

	function isOk(){return $this->ok;}

	/**
	 * @desc Check if daemon is running
	 *
	 * @return boolean
	 */
    function isAlive(){
    	if(!$this->ok) return false;		// Inexisting daemon
    	if(!$this->definition['maxUnrespondingTime']) return true;	// No check

    	// Get last ping
    	@list($t,$mem,$startTime)=cfMGetVar('daemonStatus'.$this->name);
    	return $t && (time()-$t<$this->definition['maxUnrespondingTime']);
	}

	/**
	 * @desc Return daemon's lifetime
	 *
	 * @return unknown
	 */
	function lifetime(){return $this->definition['lifetime'];}

	/**
	 * @desc Start daemon
	 *
	 */
	function start(){
		if(!$this->ok) return false;
		cfPHPAsyncScript('/local/daemons/'.$this->name.'.php',200);
		return true;
	}

	/**
	 * @desc Restart daemon
	 *
	 */
	function restart(){$this->start();}
}




/**
 * Daemon object for use in daemon scripts
 *
 */
class WDaemonInstance{
	static $name;		// Daemon name
	static $loopNb=0;	// Loop number
	static $lifetime=0;	// Max lifetime, in seconds
	static $startTime=0;// Daemon start timestamp
	static $crons=array();// Array of cron jobs

	/**
	 * @desc Initialize daemon and script
	 *
	 */
	static function init(){
		set_time_limit(0);
		ignore_user_abort(true);

		// Restrict access to localhost
		if($_SERVER['REMOTE_ADDR']!=='127.0.0.1') die('Unauthorized access');
		if(!cfCmpLeft($_SERVER['SCRIPT_NAME'],'/local/daemons/')) die('Not a daemon');

		// Start thread logging even if not enabled, as thread debug activation is manual and cannot be triggered before some daemons start
		cfThreadsDebugInit();

		// Set daemon's name
		self::$name=cfFileWithoutExtension(basename($_SERVER['SCRIPT_NAME']));

		// Debug
		if(isset($_GET['stop'])){
			cfMUnsetVar('daemonPID'.self::$name,0);
			exit;
		}

		// Set daemon's new "PID", used to ensure a single instance of daemon is running
		cfMSetVar('daemonPID'.self::$name, $_ENV['weezoScriptId']);

		// Set start time, used by ping.php to periodically restart daemon
		self::$startTime=time();

		// Set lifetime
		self::$lifetime=WDaemons::daemon(self::$name)->lifetime();

		// Do a 1st ping
		self::ping();

		//cfDebugSingle('Start "'.self::$name.'" daemon with PID:'.$_ENV['weezoScriptId']);
		echo('Start "'.self::$name.'" daemon with PID:'.$_ENV['weezoScriptId']);
	}

	/**
	 * @desc Restart this daemon
	 *
	 */
	static function restart(){cfPHPAsyncScript('/local/daemons/'.self::$name.'.php',50);}

	/**
	 * @desc Indicate daemon is still alive by setting daemonStatus[daemonName] to current time/memory footprint
	 *
	 */
	static function ping(){
		cfMSetVar('daemonStatus'.self::$name, array(time(),memory_get_usage(true),self::$startTime));

		// Debug: update memory usage
		cfThreadsDebugUpdate();
	}

	/**
	 * @desc Add a cron job (function or script call) to daemon
	 *
	 * @param string $name
	 * @param string $functionOrURL: cron job: function name or uri
	 * @param integer $interval: number of seconds between 2 calls
	 * @param integer $offset: time (in seconds) of 1st call, used not to call all jobs in the same loop
	 */
	static function addCron($name, $functionOrURL, $interval,$offset='auto'){
		if($offset==='auto') $offset=count(self::$crons); elseif ($offset<0) $offset+=$interval;
		if($offset<0 || $offset>=$interval) $offset=0;

		self::$crons[$name]=array('functionOrURL'=>$functionOrURL,'interval'=>$interval,'offset'=>$offset);
	}

	/**
	 * @desc Main loop
	 */
	static function loop(){

		while (true) {
			// If another instance of this daemon has been launched, terminate
			if(cfMGetVar('daemonPID'.self::$name)!==$_ENV['weezoScriptId']) {
				//cfDebugSingle(self::$name.' daemon exits because another instance has started');
				exit;
			}

			// If daemon is too old, restart
			if(self::$lifetime && time()-self::$startTime>self::$lifetime) {
				//cfDebugSingle(self::$name.' daemon restarts');
				self::restart();
				exit;
			}

			self::ping();
			self::$loopNb++;

			// Browse jobs to launch / call
			foreach(self::$crons as $name=>$cron) {
				if((self::$loopNb % $cron['interval'])==$cron['offset']) {
					//cfDebugSingle(self::$name.' daemon starts job: '.$cron['functionOrURL']);
					if(function_exists($cron['functionOrURL'])) call_user_func($cron['functionOrURL']); else cfPHPAsyncScript($cron['functionOrURL'],100);
				}
			}

			sleep(1);

			// Monitor memory usage and activity
			cfThreadsDebugUpdate();
		}
	}
}
